@@ -49,6 +49,9 @@ typedef struct {
 G_LOCK_DEFINE (used_devs);
 static GHashTable *used_devs = NULL;
 
+static gboolean is_obex_device (Connection *conn, 
+				const char *dev,
+				const char *name);
 
 static Connection *
 get_gwcond_connection (void)
@@ -420,78 +423,267 @@ om_dbus_disconnect_dev (const gchar *dev)
 	connection_free (conn);
 }
 
-
-#ifdef OBEX_PROGRESS
-
-#define NOTIFY_SIGNAL_RULE \
-	"type='signal',interface='com.nokia.ObexProgress',member='Cancel'"
-
-static DBusHandlerResult
-notify_message_filter (DBusConnection *dbus_conn,
-		       DBusMessage    *message,
-		       gpointer        user_data)
+static void
+om_append_paired_devices (Connection   *conn, 
+			  DBusMessage  *msg,
+                          const char   *devname, 
+			  GList       **list)
 {
-	CancelNotifyFunc  func;
-	gchar            *str;
-	GnomeVFSURI      *uri;
+        DBusMessageIter diter;
+	DBusMessageIter dsub;
 
-	func = (CancelNotifyFunc) user_data;
-
-	if (dbus_message_is_signal (message,
-				    "com.nokia.ObexProgress",
-				    "Cancel")) {
-		if (!dbus_message_get_args (message, NULL,
-					    DBUS_TYPE_STRING, &str,
-					    DBUS_TYPE_INVALID)) {
-			return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-		}
-		
-		uri = gnome_vfs_uri_new (str);
-
-		if (!uri) {
-			return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-		}
-
-		/* This is called from the main thread. */
-		
-		(* func) (uri);
-		gnome_vfs_uri_unref (uri);
+        if (!dbus_message_iter_init (msg, &diter)) {
+		return;
 	}
 
-	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	dbus_message_iter_recurse (&diter, &dsub);
+	
+	do { 
+		/* Add the entry to the list. */
+		GnomeVFSFileInfo *info;
+		char             *remote_devname;
+		
+		if (dbus_message_iter_get_arg_type (&dsub) != DBUS_TYPE_STRING) {
+			continue;
+		}
+		
+		dbus_message_iter_get_basic (&dsub, &remote_devname);
+		
+		if (!is_obex_device (conn, devname, (const char*) remote_devname)) {
+			continue;
+		}
+		
+		
+		info = gnome_vfs_file_info_new ();
+		
+		if (!info) {
+			return;
+		}
+		
+		info->flags |= GNOME_VFS_FILE_FLAGS_SYMLINK;
+		
+		info->valid_fields = 
+			GNOME_VFS_FILE_INFO_FIELDS_TYPE |
+			GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS |
+			GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE |
+			GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME;
+		
+		info->name = g_strdup_printf ("[%s]", remote_devname);
+		info->type = GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK;
+		info->permissions = 
+			GNOME_VFS_PERM_USER_READ |
+			GNOME_VFS_PERM_GROUP_READ |
+			GNOME_VFS_PERM_OTHER_READ;
+		
+		info->uid = 0;
+		info->gid = 0;
+		info->mime_type = g_strdup ("x-directory/normal");
+
+		info->symlink_name = g_strdup_printf ("obex://[%s]", remote_devname);
+
+		/*g_print ("added name: %s, symlink name: %s\n", info->name, info->symlink_name);*/
+		
+		if (!info->symlink_name) {
+			/* Extra caution. */
+			gnome_vfs_file_info_unref (info);
+			continue;
+		}
+
+		*list = g_list_append (*list, info);
+	} while (dbus_message_iter_next (&dsub));
+}
+
+/* Leave this in for easy testing. */
+#if 0
+static GList *
+append_fake_device (GList *list, const gchar *bda)
+{
+	GnomeVFSFileInfo *info;
+	
+	info = gnome_vfs_file_info_new ();
+	
+	info->valid_fields = 
+		GNOME_VFS_FILE_INFO_FIELDS_TYPE |
+		GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS |
+		GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE |
+		GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME;
+	
+	info->flags |= GNOME_VFS_FILE_FLAGS_SYMLINK;
+	
+	info->name = g_strdup_printf ("[%s]", bda);
+	info->type = GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK;
+	info->permissions = 
+		GNOME_VFS_PERM_USER_READ |
+		GNOME_VFS_PERM_GROUP_READ |
+		GNOME_VFS_PERM_OTHER_READ;
+	info->symlink_name = g_strdup_printf ("obex://[%s]", bda);
+		
+	info->uid = 0;
+	info->gid = 0;
+	info->mime_type = g_strdup ("x-directory/normal");
+
+	/*g_print ("added fake: %s %s\n", info->name, info->symlink_name);*/
+
+	return g_list_append (list, info);
+}
+#endif
+
+GList *
+om_dbus_get_dev_list (void)
+{
+	Connection      *conn;
+	DBusMessage     *msg;
+	DBusMessage     *ret1;
+        DBusMessageIter  iter;
+	DBusError        error;
+	GList           *devlist = NULL;
+
+#if 0
+	if (0) {
+		devlist = append_fake_device (devlist, "foo");
+		devlist = append_fake_device (devlist, "bar");
+		
+		return devlist;
+	}
+#endif
+	
+	conn = get_gwcond_connection ();
+	if (!conn) {
+		return NULL;
+	}
+
+	msg = dbus_message_new_method_call ("org.bluez", "/org/bluez",
+                                            "org.bluez.Manager",
+                                            "ListAdapters");
+
+	if (!msg) {
+		connection_free (conn);
+
+		return NULL;
+	}
+
+	dbus_error_init (&error);
+
+	ret1 = dbus_connection_send_with_reply_and_block (conn->dbus_conn,
+                                                          msg, -1, &error);
+	dbus_message_unref (msg);
+
+	if (dbus_error_is_set (&error)) {
+		dbus_error_free (&error);
+		connection_free (conn);
+		return NULL;
+	}
+
+	if (dbus_message_iter_init (ret1, &iter)) {
+                DBusMessageIter sub;
+
+		dbus_message_iter_recurse (&iter, &sub);
+		
+		/* Go through each entry (device) and get each paired device
+		 * from the entry.
+		 */
+		do {
+	                char *devname;
+                        DBusMessage *ret2;
+
+			dbus_message_iter_get_basic (&sub, &devname);
+
+			msg = dbus_message_new_method_call ("org.bluez",
+							    devname, 
+							    "org.bluez.Adapter", 
+							    "ListBondings");
+
+			if (!msg) {
+				dbus_message_unref (ret1);
+				connection_free (conn);
+				return NULL;
+                        }
+
+			dbus_error_init (&error);
+
+			ret2 = dbus_connection_send_with_reply_and_block (conn->dbus_conn, 
+									  msg, 
+									  -1, 
+									  &error);
+			dbus_message_unref (msg);
+
+			if (dbus_error_is_set (&error)) {
+				dbus_error_free (&error);
+
+				continue;
+			}
+
+
+                        om_append_paired_devices (conn, 
+						  ret2, 
+						  devname,
+                                                  &devlist);
+
+			dbus_message_unref (ret2);
+			
+		} while (dbus_message_iter_next (&sub));
+	}
+
+	dbus_message_unref (ret1);
+
+	connection_free (conn);
+
+	return devlist;
 }
 
 static gboolean
-_om_dbus_init_cancel_monitor (CancelNotifyFunc cancel_func)
+is_obex_device (Connection *conn,
+		const char *dev, 
+		const char *name)
 {
-	DBusConnection *conn;
-	DBusError       error;
+	DBusMessage      *msg, *ret;
+	DBusMessageIter  iter, sub;
+	DBusError        error;
+	gboolean         is_obex = FALSE;
+	char            *class_name;
+
+	msg = dbus_message_new_method_call ("org.bluez",
+					    dev,
+					    "org.bluez.Adapter",
+					    "GetRemoteServiceClasses");
+
+	dbus_message_iter_init_append (msg, &iter);
+
+	dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &name);
 
-	/* Get our own connection to make threads work. */
 	dbus_error_init (&error);
-        conn = dbus_bus_get_private (DBUS_BUS_SESSION, &error);	
-	if (!conn) {
-		g_printerr ("Failed to connect to the D-BUS daemon: %s", error.message);
-		
+	ret = dbus_connection_send_with_reply_and_block (conn->dbus_conn,
+							 msg, 
+							 -1, 
+							 &error);
+	dbus_message_unref (msg);
+
+	if (dbus_error_is_set (&error)) {
 		dbus_error_free (&error);
 		return FALSE;
 	}
 
-	dbus_bus_add_match (conn, NOTIFY_SIGNAL_RULE, NULL);
-	dbus_connection_add_filter (conn, notify_message_filter, 
-				    cancel_func, NULL);
+	if (dbus_message_iter_init (ret, &iter)) {
+		dbus_message_iter_recurse (&iter, &sub);
 
-	dbus_connection_setup_with_g_main (conn, NULL);
+		do {
+			if (dbus_message_iter_get_arg_type (&sub) != DBUS_TYPE_STRING) {
+				continue;
+			}
 
-	return FALSE;
+			dbus_message_iter_get_basic (&sub, &class_name);
+
+			if (!strcmp (class_name, "object transfer")) {
+				is_obex = TRUE;
+				break;
+			}
+
+
+		} while (dbus_message_iter_next (&sub));
+	}
+
+	dbus_message_unref (ret);
+
+	return is_obex;
 }
 
-void
-om_dbus_init_cancel_monitor (CancelNotifyFunc cancel_func)
-{
-	/* This is called from any thread so we need to set up the connection in
-	 * an idle.
-	 */
-	g_idle_add ((GSourceFunc) _om_dbus_init_cancel_monitor, cancel_func);
-}
-#endif
